from __future__ import annotations
from typing import Type, Dict
from pydantic import BaseModel, Field, create_model

from .base_factory import ContractFactoryBase


class PydanticTargetFactory(ContractFactoryBase):
    """
    Строит класс TargetRow с целевыми признаками модели.
    """

    def build(self) -> Type[BaseModel]:
        fields: Dict[str, Type] = self.targets
        return create_model(
            f"{self.prefix}_TargetRow",
            __base__=BaseModel,
            __module__=__name__,
            **{k: (t, Field(...)) for k, t in fields.items()},
        )



from __future__ import annotations
from typing import Type, Dict
from pydantic import BaseModel, Field, create_model
from .base_factory import ContractFactoryBase


class PydanticFeatureFactory(ContractFactoryBase):
    """
    Строит класс FeatureRow со входными признаками модели.
    """

    def build(self) -> Type[BaseModel]:
        fields: Dict[str, Type] = self.inputs
        return create_model(
            f"{self.prefix}_FeatureRow",
            __base__=BaseModel,
            __module__=__name__,
            **{k: (t, Field(...)) for k, t in fields.items()},
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from pathlib import Path
from os import PathLike
from typing import Type
from .schema import ContractSchema
from .encoders import NaInfEncoder, DEFAULT_ENCODER


class ContractFactoryBase(ABC):
    """
    Базовый шаблон для класов фабрик.

    Читаем YAML (один раз) и предоставляем содержимое секций
    inputs и targets наследникам.

    Метод build() для генерации динамического класса.
    """

    def __init__(
        self,
        yaml_path: str | PathLike[str],
        *,
        prefix: str | None = None,
        encoder: NaInfEncoder = DEFAULT_ENCODER,
    ) -> None:
        self.schema  = ContractSchema(yaml_path)
        self.prefix  = prefix or Path(yaml_path).stem
        self.encoder = encoder

    @property
    def inputs(self):  return self.schema.inputs

    @property
    def targets(self): return self.schema.targets

    @abstractmethod
    def build(self) -> Type: ...



from ml_dataset_contract_tte import (
    TteFeatureRow,
    TteTargetRow,
    TtePredictRequest,
)


sample_row__input = {
    "expanding_tte_mean": 1.1,
    "tte_lag_1": 1.2,
}
sample_row__target = {"tte": 1.3}

feature_row = TteFeatureRow(**sample_row__input)
print(f"{feature_row=}")

target_row = TteTargetRow(**sample_row__target)
print(f"{target_row=}")

request = TtePredictRequest(rows=[feature_row])
payload = request.to_split_json()
print(f"{payload}")



from __future__ import annotations
from pathlib import Path
from typing import Dict, Type, Tuple
import yaml

_PY_TYPES: Dict[str, Type] = {
    "float": float,
    "int":   int,
    "str":   str,
    "bool":  bool,
}

class ContractSchema:
    """
    Хранит описание разделов inputs и targets из yaml-файла.
    """

    def __init__(self, yaml_path: str | "os.PathLike[str]") -> None:
        self.path = Path(yaml_path)
        self._inputs, self._targets = self._load()

    @property
    def inputs(self) -> Dict[str, Type]:
        return self._inputs

    @property
    def targets(self) -> Dict[str, Type]:
        return self._targets

    def _load(self) -> Tuple[Dict[str, Type], Dict[str, Type]]:
        with self.path.open(encoding="utf-8") as f:
            raw = yaml.safe_load(f)

        if not {"inputs", "targets"} <= raw.keys():
            raise KeyError(
                "YAML must contain 'inputs' and 'targets' sections")

        return self._convert(raw["inputs"]), self._convert(raw["targets"])

    @staticmethod
    def _convert(raw_map: Dict[str, str]) -> Dict[str, Type]:
        unknown = set(raw_map.values()) - _PY_TYPES.keys()
        if unknown:
            raise ValueError(f"Unknown types in YAML: {unknown}")
        return {k: _PY_TYPES[v] for k, v in raw_map.items()}



import math
from typing import Any


class NaInfEncoder:
    """
    Преобразует:
      - NaN  -> None
      - +Inf -> inf_value
      - -Inf -> -inf_value
    """

    def __init__(self, inf_value: float = 9999):
        self.inf_value = inf_value

    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, float):
            if math.isnan(obj):
                return None
            if math.isinf(obj):
                return self.inf_value if obj > 0 else -self.inf_value
            return obj
        if isinstance(obj, (list, tuple)):
            return [self(x) for x in obj]
        if isinstance(obj, dict):
            return {k: self(v) for k, v in obj.items()}
        return obj


DEFAULT_ENCODER: NaInfEncoder = NaInfEncoder()



from importlib.resources import files

from ml_dataset_contract.runtime import (
    PydanticFeatureFactory,
    PydanticTargetFactory,
    PydanticRequestFactory,
)

_YAML = files(__name__).joinpath("dataset.yml")

TteFeatureRow = PydanticFeatureFactory(_YAML, prefix="Tte").build()
TteTargetRow  = PydanticTargetFactory(_YAML, prefix="Tte").build()
TtePredictRequest = PydanticRequestFactory(
    _YAML, prefix="Tte", feature_cls=TteFeatureRow
).build()

__all__ = [
    "TteFeatureRow", "TteTargetRow", "TtePredictRequest", "_YAML"
]


from __future__ import annotations
from typing import Type, List, Any, Dict
from pydantic import BaseModel, model_validator
from .base_factory import ContractFactoryBase
from .feature_factory import PydanticFeatureFactory
from .encoders import NaInfEncoder


class PydanticRequestFactory(ContractFactoryBase):
    """
    Строит Dto-класс PredictRequest с полями для REST API
    клиента модели (mlflow serve).

    Формат запроса, который по умолчанию понимимает Mlflow Serve имеет
    вид:
    ```json
    {
        "dataframe_split": {
            "columns": [...],
            "data": [[...], ...]
        }
    }
    ```
    """

    def __init__(self, *args, feature_cls: Type[BaseModel], **kw) -> None:
        super().__init__(*args, **kw)
        self.feature_cls = feature_cls
        self._cols = list(self.inputs)

    def build(self) -> Type[BaseModel]:
        cols = self._cols
        enc  = self.encoder
        FeatureRow = self.feature_cls

        class PredictRequest(BaseModel):
            rows: List[FeatureRow]
            model_config = {"json_encoders": {float: enc}}

            @model_validator(mode="after")
            def _no_empty(self):
                if not self.rows:
                    raise ValueError("rows can't be empty")
                return self

            def to_split_json(
                self, *, encoder: NaInfEncoder | None = None
            ) -> Dict[str, Any]:
                e = encoder or enc
                data = [[e(getattr(r, c)) for c in cols] for r in self.rows]
                return {"dataframe_split": {"columns": cols, "data": data}}

        PredictRequest.__name__ = f"{self.prefix}_PredictRequest"
        return PredictRequest



from pathlib import Path
from setuptools import setup, find_namespace_packages

ROOT = Path(__file__).parent

setup(
    name="ml_dataset_contract_core",
    description="Runtime core for ML dataset contracts.",
    long_description=(
        ROOT / "README.md"
    ).read_text(encoding="utf-8") if (ROOT / "README.md").exists() else "",
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    license="MIT",
    author="Nikolskii D.N.",
    url="https://github.com/nikolskydn/ml_dataset_contract_core",
    use_scm_version={
        "version_scheme": "no-guess-dev",
        "local_scheme":  "no-local-version",
        "fallback_version": "0.0.0",
    },
    setup_requires=["setuptools_scm[toml]>=8"],
    # --------------------------------------------
    package_dir={"": "src"},
    packages=find_namespace_packages(
        "src",
        include=["ml_dataset_contract.runtime*"],
    ),
    install_requires=[
        "pydantic>=2.6",
        "pyyaml>=6.0",
    ],
    include_package_data=False,
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Framework :: Pydantic",
        "Operating System :: OS Independent",
    ],
)



from .schema import ContractSchema
from .base_factory import ContractFactoryBase
from .feature_factory import PydanticFeatureFactory
from .target_factory import PydanticTargetFactory
from .request_factory import PydanticRequestFactory
from .encoders import NaInfEncoder, DEFAULT_ENCODER


__all__ = [
    "ContractSchema",
    "ContractFactoryBase",
    "PydanticFeatureFactory",
    "PydanticTargetFactory",
    "PydanticRequestFactory",
    "NaInfEncoder",
    "DEFAULT_ENCODER",
]
